${eCAP_AV_SDK_BOILERPLATE}

#include "Adapter.h"
#include "Answers.h"
#include "Service.h"
#include "Xaction.h"
#include "${eCAP_AV_SDK_CLASS_NAME}.h"
#include "Debugger.h"
#include "Gadgets.h"
#include <libecap/common/errors.h>
#include <libecap/common/named_values.h>
#include <iostream>
#include <sstream>
#include <cstring>
#include <cstdio>
#include <climits>
#include <cerrno>
#include <stdlib.h>

static const time_t timeNone = static_cast<time_t>(-1);

// Time between antivirus database update checks
// Is it better to update externally instead?
//    Yes, but would external updates affect existing engine??
// 0 = on every access; timeNone = never
const time_t dbUpdateGap = 60; // in seconds

// default staging filename template
static const std::string TmpFileNameTemplateDefault =
	"/tmp/${eCAP_AV_SDK_SHORT_NAME}XXXXXX"; // TODO: use $TEMP


Adapter::Service::Service(const std::string &aMode):
	mode(aMode),
	vbAccumulationLimit(0),
	blockOnError(false),
	scanAsynchronously(AllowAsyncScans()),
	answers(0),
	lastDbUpdate(0)
{
	if (AllowAsyncScans())
		answers = new Answers;
}

Adapter::Service::~Service()
{
	if (answers)
		answers->abandon();
}

std::string Adapter::Service::uri() const
{
	return "${eCAP_AV_SDK_SERVICE_URI}?mode=" + mode;
}

std::string Adapter::Service::tag() const
{
	return PACKAGE_VERSION;
}

void Adapter::Service::describe(std::ostream &os) const
{
	os << PACKAGE_NAME << " v" << PACKAGE_VERSION;
}

// pthread support test (to avoid spreading ifdefs around the code)
bool Adapter::Service::AllowAsyncScans() {
#ifdef HAVE_PTHREAD
	return true;
#else
	return false;
#endif
}

bool Adapter::Service::makesAsyncXactions() const
{
	if (scanAsynchronously)
		return true;

	// Our scanAsynchronously may become false during reconfiguration
	// while we still have pending asynchronous transactions and, hence,
	// still need the host to treat us as an async service. The first answers
	// user is this Service, so we only interested in higher user counts.
	return AllowAsyncScans() && answers && (answers->users() > 1);
}

namespace Adapter {
class Cfgtor: public libecap::NamedValueVisitor {
	public:
		Cfgtor(Service &aSvc): svc(aSvc) {}
		virtual void visit(const libecap::Name &name, const libecap::Area &value) {
			svc.setOne(name, value);
		}
		Service &svc;
};
} // namespace Adapter

void Adapter::Service::configure(const Options &cfg) {
	// a workaround for host application that should have called reconfigure()
	if (scanner) {
		reconfigure(cfg);
		return;
	}

	setAll(cfg);

	// create an antivirus scanner; TODO: should some services share instances?
	Must(!scanner);
	scanner.reset(new ${eCAP_AV_SDK_CLASS_NAME});
	scanner->configure(cfg);

	checkpoint();
}

// we may be called from configure() if a buggy host does not call us directly
void Adapter::Service::reconfigure(const Options &newCfg)
{
	setAll(newCfg);
	Must(scanner);
	scanner->reconfigure(newCfg);
	checkpoint();
}

void Adapter::Service::setAll(const Options &cfg) {
	tmpFileNameTemplate = TmpFileNameTemplateDefault;

	Cfgtor cfgtor(*this);
	cfg.visitEachOption(cfgtor);

	checkStagingDir();

	Debugger(ilDebug|flApplication) << "async=" << scanAsynchronously;
}

void Adapter::Service::setOne(const libecap::Name &name, const libecap::Area &valArea) {
	const std::string value = valArea.toString();
	if (name == "on_error")
		setOnError(value);
	else
	if (name == "staging_dir")
		setTmpDir(value);
	else
	if (name == "huge_size")
		setAccumulationLimit(value);
	else
	if (name == "debug")
		; // the scanner handles that (TODO: ask the scanner instead)
	else
	if (name == "async")
		setAsyncScans(value);
	else
	if (name.assignedHostId())
		; // skip host-specific options
	else
		throw libecap::TextException("${eCAP_AV_SDK_SHORT_NAME}: "
			"unsupported adapter configuration parameter: " + name.image());
}

void Adapter::Service::setOnError(const std::string &value) {
	// default is not to block
	if (value == "block")
		blockOnError = true;
	else
	if (value == "allow")
		blockOnError = false;
	else
		throw libecap::TextException("${eCAP_AV_SDK_SHORT_NAME}: unsupported on_error config "
			"value (" + uri() + "): " + value);
}

void Adapter::Service::setTmpDir(const std::string &prefix) {
	std::string temp = prefix;
	if (temp.empty() || temp == "default")
		temp = TmpFileNameTemplateDefault;
	if (temp.rfind('X') != temp.size()-1)
		temp += "XXXXXX";
	tmpFileNameTemplate = temp;
}

void Adapter::Service::checkStagingDir() {
	std::string tmpFileName = tmpFileNameTemplate;
	if (FILE *file = createTempFile(tmpFileName)) {
		// we do not check problems with syscalls below because Xactions do not
		fclose(file);
		remove(tmpFileName.c_str());
	}
}

void Adapter::Service::setAccumulationLimit(const std::string &value) {
	if (value == "none") {
		vbAccumulationLimit = 0; // no limit
		return;
	}

	std::istringstream input(value);
	uint64_t size;
	if (input >> size) {
		vbAccumulationLimit = size;
		return;
	}
	const std::string msg = "invalid huge_size parameter value: " + value;
	throw libecap::TextException(msg);
}

void Adapter::Service::setAsyncScans(const std::string &value) {
	if (value == "no") {
		scanAsynchronously = false;
	} else if (value == "yes") {
		if (!AllowAsyncScans()) {
			const std::string msg = "Cannot honor async=yes "
				"without POSIX threads (pthreads) support.";
			throw libecap::TextException(msg);
		}
		scanAsynchronously = true;
	} else {
		const std::string msg = "Unsupported value in async=" +
			value + ". Expected 'yes' or 'no' value.";
		throw libecap::TextException(msg);
	}
}

void Adapter::Service::start()
{
	Must(tmpFileNameTemplate.size() > 0); // we were successfully configured
	libecap::adapter::Service::start();
}

void Adapter::Service::stop()
{
	libecap::adapter::Service::stop();
}

void Adapter::Service::retire()
{
	libecap::adapter::Service::retire();
}

bool Adapter::Service::wantsUrl(const char *url) const
{
	return true; // no-op is applied to all messages
}

Adapter::Service::MadeXactionPointer
Adapter::Service::makeXaction(libecap::host::Xaction *hostx)
{
	checkpoint();
	const libecap::shared_ptr<Service> s = std::tr1::static_pointer_cast<Service>(self);
	XactionPointer x(new Xaction(s, hostx));
	x->self = x;
	return x;
}

void Adapter::Service::suspend(timeval & timeout)
{
	Must(answers);
	if (answers->empty()) {
		// limit wait to maxUsec
		// TODO: there should be no limit when no answers are expected
		const int maxUsec = 300*1000; 
		if (timeout.tv_sec || timeout.tv_usec > maxUsec) {
			timeout.tv_sec = 0;
			timeout.tv_usec = maxUsec;
		}
	} else {
		// no time to wait at all -- at least one answer is ready already
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
	}
}

void Adapter::Service::resume()
{
	Must(answers);
	while (Answer *answer = answers->get()) {
		if (MyAnswer *ma = dynamic_cast<MyAnswer*>(answer)) {
			ma->xaction->tellHostToResume(answer);
		} else {
			Debugger(ilCritical|flXaction) << "internal error: " <<
				"Adapter::Service::resume() got wrong scan answer object: " <<
				answer->fileName << ".";
			delete answer;
		}
	}
}

void Adapter::Service::checkpoint()
{
	if (dbUpdateGap == timeNone)
		return; // no updates configured

	if (time(0) < lastDbUpdate + dbUpdateGap)
		scanner->update();

	// we enforce the time gap _between_ updates so that even relatively long
	// updates do not lead to gap-free uptates
	lastDbUpdate = time(0); 
}
